@@ -1,28 +1,24 @@
 #!/usr/bin/env python
-import sys
-import os
-import logging
-import types
 import io
-import tarfile
-import zipfile
+import json
+import logging
 import marshal
+import os
+import re
+import sys
+import tarfile
+import types
+import zipfile
 from contextlib import contextmanager
-
-LEGACY = (sys.version_info.major == 2)
-if LEGACY:
-    import imp
-    from ConfigParser import ConfigParser, NoSectionError
-    from urllib2 import urlopen, Request, HTTPError, URLError, ProxyHandler, build_opener, install_opener
-else:
-    from configparser import ConfigParser, NoSectionError
-    from urllib.request import urlopen, Request, ProxyHandler, build_opener, install_opener
-    from urllib.error import HTTPError, URLError
+from configparser import ConfigParser, NoSectionError
+from urllib.error import HTTPError, URLError
+from urllib.request import (ProxyHandler, Request, build_opener,
+                            install_opener, urlopen)
 
 # ====================== Metadata ======================
 
 __author__ = 'John Torakis - operatorequals'
-__version__ = '1.1.0'
+__version__ = '1.2.0'
 __github__ = 'https://github.com/operatorequals/httpimport'
 
 # ====================== Constants ======================
@@ -57,6 +53,18 @@ headers:
     X-HttpImport-Version: {version}
     X-HttpImport-Project: {homepage}
 
+# PyPI specific:
+# A multi-line with 'requirements.txt' syntax
+requirements:
+
+# Filepath of a 'requirements.txt' file
+requirements-file:
+
+# A multi-line with (module_name, pypi_project) tuples
+# e.g.:
+#   bs4: beautifulsoup4
+project-names:
+
 ### Not Implemented ###
 # allow-compiled: no
 
@@ -75,10 +83,7 @@ _DEFAULT_INI_CONFIG_FILENAME = __HOME_DIR + os.sep + ".httpimport.ini"
 _DEFAULT_INI_CONFIG_DIR_NAME = __HOME_DIR + os.sep + ".httpimport"
 
 CONFIG = ConfigParser()
-if LEGACY:
-    CONFIG.readfp(io.BytesIO(_DEFAULT_INI_CONFIG))
-else:
-    CONFIG.read_string(_DEFAULT_INI_CONFIG)
+CONFIG.read_string(_DEFAULT_INI_CONFIG)
 
 # ====================== Logging ======================
 
@@ -96,7 +101,7 @@ logger.addHandler(log_handler)
 
 
 def http(url, headers={}, method='GET', proxy=None):
-    """ Wraps Python2/3 HTTP calls and ensures cross-compatibility
+    """ Wraps HTTP/S calls in one place
 
     Args:
         url (str):
@@ -107,12 +112,7 @@ def http(url, headers={}, method='GET', proxy=None):
     Returns:
         dict: A dict containing 'code', 'headers', 'body' of HTTP response
     """
-
-    if LEGACY:
-        req = Request(url, headers=headers)
-        req.get_method = lambda: method.upper()
-    else:
-        req = Request(url, headers=headers, method=method.upper())
+    req = Request(url, headers=headers, method=method.upper())
 
     if proxy:
         scheme = proxy.split(':', 1)[0]
@@ -124,11 +124,7 @@ def http(url, headers={}, method='GET', proxy=None):
 
     try:
         resp = _urlopen(req)
-        try:  # Python2 Approach
-            headers = {k.lower(): v for k, v in resp.headers.dict.items()}
-        except AttributeError:  # Python3 Approach
-            headers = {k.lower(): v for k, v in resp.getheaders()}
-
+        headers = {k.lower(): v for k, v in resp.getheaders()}
         return {'code': resp.code, 'body': resp.read(), 'headers': headers}
     except HTTPError as he:
         return {'code': he.code, 'body': b'', 'headers': {}}
@@ -140,7 +136,7 @@ def _isHTTPS(url):
     return url.startswith('https://')
 
 
-def _get_url_options(url):
+def _get_options(url):
     if url in CONFIG.sections():
         return dict(CONFIG.items(url))
     else:
@@ -233,7 +229,53 @@ def _retrieve_compiled(content):  # <== Not Used Yet
 
     raise ValueError("[!] Not possible to unmarshal '.pyc' file")
 
-# ====================== Importer Class ======================
+
+def _create_pypi_url(
+        module_name,
+        version=None,
+        allowed_dists=[
+            'bdist_wheel',
+            'sdist'],
+        pypi_url="https://pypi.org/pypi/%s/json"):
+    """ Returns the URL of a PyPI distribution of a module.
+The Download URL is acquired by directly querying the PyPI API:
+https://warehouse.pypa.io/api-reference/json.html
+    """
+    url = pypi_url % module_name
+    logger.debug("[+] Querying PyPI URL '%s'" % url)
+    try:
+        raw_response = http(url)
+        pypi_response = json.loads(raw_response['body'])
+    except json.decoder.JSONDecodeError:
+        raise ModuleNotFoundError(
+            "PyPI API did not respond with JSON for '%s'. HTTP Status Code: %d" %
+            (module_name, raw_response['code']))
+    if version is None:
+        version = pypi_response['info']['version']
+    if version not in pypi_response['releases']:
+        raise KeyError(
+            "Version '%s' not available for module %s" %
+            version, module_name)
+    release = pypi_response['releases'][version]
+    logger.info(
+        "[+] Version '%s' found for module '%s'" %
+        (version, module_name))
+    for package in release:
+        if 'url' not in package:
+            logger.info(
+                "[-] Version '%s' is an empty release for module '%s'" %
+                version, module_name)
+            continue
+        if package['packagetype'] in allowed_dists:
+            logger.info(
+                "[+] Version '%s' release available in %s" %
+                (version, allowed_dists))
+            return package['url']
+    raise KeyError(
+        "No allowed release type found for %s==%s. Allowed release types: %s" %
+        (module_name, version, allowed_dists))
+
+# ====================== Importer Classes ======================
 
 
 class HttpImporter(object):
@@ -256,7 +298,7 @@ class HttpImporter(object):
             zip_pwd=b'',
             headers={},
             proxy=None,
-            allow_plaintext=False):
+            allow_plaintext=False, **kw):
         # remove trailing '/' from URL parameter
         self.url = url if not url.endswith('/') else url[:-1]
         self.modules = {}
@@ -266,8 +308,7 @@ class HttpImporter(object):
                 "[-] Using HTTP URLs (%s) with 'httpimport' is a security hazard!" %
                 (url))
             if not (allow_plaintext or INSECURE):
-                logger.error(""" [*]
-Using plaintext protocols needs to be enabled through 'INSECURE' global or explicitly allowed through 'allow-plaintext'!
+                logger.error("""[*] Using plaintext protocols needs to be enabled through 'INSECURE' global or explicitly allowed through 'allow-plaintext'!
                 """)
                 raise ImportError(
                     "[-] HTTP used while plaintext is not allowed")
@@ -327,7 +368,7 @@ Using plaintext protocols needs to be enabled through 'INSECURE' global or expli
                         (path))
                     self.modules[fullname] = {}
                     self.modules[fullname]['content'] = content
-                    self.modules[fullname]['filepath'] = path
+                    self.modules[fullname]['filepath'] = self.url + "#" + path
                     self.modules[fullname]['package'] = path.endswith(
                         '__init__.py')
                     return self
@@ -359,12 +400,12 @@ Using plaintext protocols needs to be enabled through 'INSECURE' global or expli
         # method (yet)
         if fullname not in self.modules:
             logger.debug(
-                "[*] Module '%s' has not been attempted before. Trying to load...")
+                "[*] Module '%s' has not been attempted before. Trying to load..." % fullname)
             # Run 'find_module' and see if it is loadable through this Importer
             # object
             if self.find_module(fullname) is not self:
                 logger.info(
-                    "[-] Module '%s' has not been found as loadable. Failing...")
+                    "[-] Module '%s' has not been found as loadable. Failing..." % fullname)
                 # If it is not loadable ('find_module' did not return 'self' but 'None'):
                 # throw error:
                 raise ImportError(
@@ -374,11 +415,7 @@ Using plaintext protocols needs to be enabled through 'INSECURE' global or expli
         logger.debug(
             "[*] Creating Python Module object for '%s'" % (fullname))
 
-        if LEGACY:
-            mod = imp.new_module(fullname)
-        else:
-            mod = types.ModuleType(fullname)
-
+        mod = types.ModuleType(fullname)
         mod.__loader__ = self
         mod.__file__ = self.modules[fullname]['filepath']
         # Set module path - get filepath and keep only the path until filename
@@ -412,7 +449,6 @@ Using plaintext protocols needs to be enabled through 'INSECURE' global or expli
                 logger.warning(
                     "[-] Module/Package '%s' cannot be imported without adding it to sys.modules. Might contain relative imports." %
                     fullname)
-
         return mod
 
     def load_module(self, fullname):
@@ -426,20 +462,111 @@ Using plaintext protocols needs to be enabled through 'INSECURE' global or expli
 
         """
         logger.info("[*] Loading module '%s' into sys.modules" % fullname)
+        mod = self._create_module(fullname)
+        return sys.modules[fullname]
 
-        if LEGACY:
-            imp.acquire_lock()
+
+class PyPIImporter(object):
+    """ The class that implements the Importer API. Uses the HttpImporter to import PyPI
+releases.
+
+    Args:
+        version_matrix (dict):
+        project_matrix (dict):
+        allowed_dists (list):
+        pypi_url (str):
+        **kw (dict): Parameters that are passed to HttpImporter objects created by this class
+     """
+
+    def __init__(
+            self,
+            url="https://pypi.org/pypi/%s/json",
+            project_matrix={},
+            version_matrix={},
+            allowed_dists=[
+                'bdist_wheel',
+                'sdist'], **kw):
+        if url is None:
+            url = 'https://pypi.org/pypi/%s/json'
+        self.url = url  # Duck Type with HttpImporter
+        self.version_matrix = version_matrix
+        self.project_matrix = project_matrix
+        self.allowed_dists = allowed_dists
+        self.module_importers = {}
+        self.kw = kw
+
+    def find_module(self, module_name, path=None):
+        logger.info(
+            "[*] Trying to find PyPI module '%s', path: '%s'" %
+            (module_name, path))
+        module_root = module_name.split('.')[0]
+        if module_root in self.module_importers:
+            return self.module_importers[module_root]
+
+        version = None
+        # Get the PyPI Project from Module name, if not available use module
+        # root
+        project_name = self.project_matrix.get(module_root, module_root)
+        version_tuple = self.version_matrix.get(project_name, (None, None))
+        # Parse version tuple ('==', '1.0.0')
+        if version_tuple[0] == '==':
+            version = version_tuple[1]
 
         try:
-            mod = self._create_module(fullname)
-            return sys.modules[fullname]
-        finally:
-            if LEGACY:
-                imp.release_lock()
+            url = _create_pypi_url(
+                project_name,
+                version=version,
+                allowed_dists=self.allowed_dists,
+                pypi_url=self.url)
+            importer = HttpImporter(url, **self.kw)
+            found = importer.find_module(module_name)
+            if found:
+                logger.info(
+                    "[+] Module '%s' can be loaded from PyPI project '%s'. URL: '%s'" %
+                    (module_name, project_name, url))
+                self.module_importers[module_root] = found
+                return found
+
+        except (KeyError, ModuleNotFoundError) as e:
+            logger.warning("[-] %s" % e)
+        # Could not load module from PyPI
+        logger.warning(
+            "[-] Module '%s' cannot be found in PyPI." %
+            module_name)
+        return None
+
+    def _create_module(self, fullname, sys_modules=True):
+        module_root = fullname.split('.')[0]
+        if module_root not in self.module_importers:
+            logger.debug(
+                "[*] Module '%s' has not been attempted before. Trying to find in PyPI..." % fullname)
+            # Run 'find_module' and see if it returns an HttpImporter
+            # object
+            if type(self.find_module(fullname)) != HttpImporter:
+                logger.info(
+                    "[-] Module '%s' has not been found in PyPI. Failing..." % fullname)
+                # If it is not loadable ('find_module' did not return HttpImporter):
+                raise ImportError(
+                    "Module '%s' cannot be loaded from PyPI" %
+                    (fullname))
+        return self.module_importers[module_root]._create_module(fullname, sys_modules)
+
+    def load_module(self, fullname):
+        logger.info(
+            "[*] Loading PyPI module '%s'" %
+            (fullname))
+        try:
+            return self._create_module(fullname)
+        except KeyError:
+            pass
+        logger.warning(
+            "[-] Module '%s' could not be imported from PyPI." %
+            fullname)
 
 
 # ====================== Feature Helpers ======================
 
+
 def __create_git_url(service, username=None, repo=None,
                      ref='master', domain=None):
     """ Function that creates a URL of a Managed Git Service. Uses the internal
@@ -465,23 +592,20 @@ def __create_git_url(service, username=None, repo=None,
 def __extract_profile_options(url=None, profile=None):
     if profile:
         # If there is a profile name set - try it
-        options = _get_url_options(profile)
+        options = _get_options(profile)
         logger.debug("[*] Profile for URL: '%s' -> %s" % (url, options))
         logger.debug(
             "[*] Profile '%s' for URL: '%s' -> %s" %
             (profile, url, options))
         if url is None:
-            url = options['url']
+            # if there is no 'url' in there too, could be a PyPI profile
+            url = options.get('url', None)
     else:
-        options = _get_url_options(url)
+        options = _get_options(url)
         logger.debug("[*] Profile for URL: '%s' -> %s" % (url, options))
 
     proxy = options['proxy-url']
-
-    # Python2/3 Bytes vs Str issue. Start with Python2 way, try Python3
-    zip_pwd = options['zip-password']
-    if not LEGACY:
-        zip_pwd = bytes(options['zip-password'], 'utf8')
+    zip_pwd = bytes(options['zip-password'], 'utf8')
 
     # Parse header dict from str lines
     headers = {
@@ -490,28 +614,57 @@ def __extract_profile_options(url=None, profile=None):
         if line
     }
 
-    # Parse Plaintext Flag
+    # Parse allow-HTTP Flag
     allow_plaintext = options['allow-plaintext'].lower() in ['true',
                                                              'yes', '1']
+    # Get PyPI requirements
+    requirements_file = options['requirements-file']
+    requirements = options['requirements']
+    # Add file requirements to requirement dict
+    if requirements_file:
+        with open(requirements_file) as req_file:
+            requirements += '\n' + req_file.read()
+
+    # Parse requirements
+    version_matrix = {}
+    for line in requirements.splitlines():
+        if not line:
+            continue
+        match = re.match(r'^([\w_-]+)\s*(([=<>]=)\s*(\d+\.\d+\.\d+))?', line)
+        if not match:
+            continue
+
+        if match.group(3, 4):
+            version_matrix[match.group(1)] = (match.group(3, 4))
+        else:
+            version_matrix[match.group(1)] = (match.group(None))
+
+    # Parse header dict from str lines
+    project_matrix = {
+        line.split(':', 1)[0].strip(): line.split(':', 1)[1].strip()
+        for line in options['project-names'].splitlines()
+        if line
+    }
+
     return {
         'headers': headers,
         'proxy': proxy,
         'url': url,
         'zip_pwd': zip_pwd,
-        'allow-plaintext': allow_plaintext}
+        'allow_plaintext': allow_plaintext,
+        'version_matrix': version_matrix,
+        'project_matrix': project_matrix,
+    }
 
 # ====================== Features ======================
 
 
 def set_profile(ini_str):
     global CONFIG
-    if LEGACY:
-        CONFIG.readfp(io.BytesIO(ini_str))
-    else:
-        CONFIG.read_string(ini_str)
+    CONFIG.read_string(ini_str)
 
 
-def add_remote_repo(url=None, profile=None):
+def add_remote_repo(url=None, profile=None, importer_class=HttpImporter):
     """ Creates an HttpImporter object and adds it to the `sys.meta_path`.
 
     Args:
@@ -522,12 +675,14 @@ def add_remote_repo(url=None, profile=None):
       HttpImporter: The `HttpImporter` object added to the `sys.meta_path`
     """
     options = __extract_profile_options(url, profile)
-    importer = HttpImporter(
+    url = options.get('url', url)
+    del options['url']
+    logger.debug(
+        "[*] Adding '%s' (profile: %s) with options: %s " %
+        (importer_class, profile, options))
+    importer = importer_class(
         url,
-        headers=options['headers'],
-        proxy=options['proxy'],
-        zip_pwd=options['zip_pwd'],
-        allow_plaintext=options['allow-plaintext'],
+        **options,
     )
     sys.meta_path.append(importer)
     return importer
@@ -561,7 +716,10 @@ def remote_repo(url=None, profile=None):
     or of an archive (supported: .zip, .tar, .tar.bz, .tar.gz, .tar.xz - Python3 only)
 
     """
-    importer = add_remote_repo(url=url, profile=profile)
+    importer = add_remote_repo(
+        url=url,
+        profile=profile,
+        importer_class=HttpImporter)
     url = importer.url
     try:
         yield
@@ -640,7 +798,7 @@ def gitlab_repo(username=None, repo=None, ref='master',
         remove_remote_repo(url)
 
 
-def load(module_name, url=None, profile=None):
+def load(module_name, url=None, profile=None, importer_class=HttpImporter):
     """ Loads a module on demand and returns it as a module object. Does NOT load it to the Namespace.
   Example:
 
@@ -650,20 +808,38 @@ def load(module_name, url=None, profile=None):
   >>>
     """
     options = __extract_profile_options(url, profile)
-    importer = HttpImporter(
+    url = options.get('url', url)
+    del options['url']
+    importer = importer_class(
         url,
-        headers=options['headers'],
-        proxy=options['proxy'],
-        zip_pwd=options['zip_pwd'],
-        allow_plaintext=options['allow-plaintext'],
+        **options,
     )
     return importer._create_module(module_name, sys_modules=False)
     raise ImportError(
         "Module '%s' cannot be imported from URL: '%s'" % (module_name, url))
 
+@contextmanager
+def pypi_repo(url='https://pypi.org/pypi/%s/json', profile=None):
+    """ Context Manager that provides remote import functionality from PyPI
+
+    Args:
+        TODO
+    """
+    importer = add_remote_repo(
+        url=url,
+        profile=profile,
+        importer_class=PyPIImporter)
+    url = importer.url
+    try:
+        yield
+    except ImportError as e:
+        raise e
+    finally:  # Always remove the added Importer from sys.meta_path
+        remove_remote_repo(url)
 
 # ====================== Runtime ======================
 
+
 # Load the catch-all configuration
 set_profile(_DEFAULT_INI_CONFIG)
 
@@ -674,7 +850,7 @@ try:
             "[*] Loading configuration from '%s'" %
             _DEFAULT_INI_CONFIG_FILENAME)
         set_profile(f.read())
-except IOError: # Not using FileNotFoundError - Python2 compatibility
+except FileNotFoundError:
     logger.info("[*] File '%s' not available." % _DEFAULT_INI_CONFIG_FILENAME)
 
 # Try to load the user config directory
@@ -683,7 +859,7 @@ try:
         with open(config_file) as f:
             logger.info("[*] Loading configuration from '%s'" % config_file)
             set_profile(f.read())
-except OSError: # Not using FileNotFoundError - Python2 compatibility
+except FileNotFoundError:
     logger.info(
         "[*] Directory '%s' not available." %
         _DEFAULT_INI_CONFIG_DIR_NAME)
